%%%%%%% MATLAB (R) script to compute TWDP %%%%%%%%%%%%%%%%%%%%%%
% *** MODIFIED TO INCLUDE AUTOMATIC OMA ESTIMATION. USES SEPARATE SCRIPT Estimate_OMA.m
% *** Three methods for estimating OMA are provided.  Select one by setting OMA_Method parameter
if ~exist('OMA_Method');    % Valid values are 0 (not automatic), 1, 2, 3
    OMA_Method  = 0;        % Defaults to behaviour in D2.2 if OMA_Method not set
end
%% TP-2 test inputs
%% Units for all optical power values must match.
 
%% The values given below for TxDataFile and MeasuredWaveformFile are
%% examples and should be replaced by actual path\filenames and values for each waveform tested.

%% Transmit data file: The transmit data sequence is one of the TWDP test patterns defined in
%% Table 686. The file format is ASCII with a single column of chronological ones and zeros with no
%% headers or footers.
TxDataFile = 'prbs9_950.txt';
%% Measured waveform. The waveform consists of exactly N samples per unit interval T, where N is the
%% oversampling rate. The waveform must be circularly shifted to align with the data sequence. The file
%% format for the measured waveform is ASCII with a single column of chronological numerical samples,
%% in optical power, with no headers or footers.
MeasuredWaveformFile = 'preproc-1207-01.txt';
 
%% OMA and steady-state ZERO power must be entered.
%% ***Automatic estimation replaces these entries
if ~OMA_Method
    MeasuredOMA        = 6.6e-4;   % Measured OMA, in optical power
    SteadyZeroPower    = 2.76e-4;  % Measured steady-state logic ZERO
end
OverSampleRate      = 16;       % Oversampling rate
%% Simulated fiber response, modeled as a set of ideal delta functions with specified amplitudes in optical
%% power and delays in nanoseconds, in rows. The three cases specified in Table 685 for the
%% comprehensive stressed receiver tests are used. The vector 'PCoefs' contains the amplitudes, and the
%% vector 'Delays' contains the delays.
FiberResp = [...
0.000000 0.072727 0.145455 0.218182
0.158 0.176 0.499 0.167
0.000 0.513 0.000 0.487
0.254 0.453 0.155 0.138];
Delays          = FiberResp(1,:)';
 
%% Program constants %%
SymbolPeriod    = 1/10.3125;        % Symbol period (ns)
EFilterBW       = 7.5;              % Front end filter bandwidth (GHz)
EqNf            = 100;              % Number of feedforward equalizer taps
EqNb            = 50;               % Number of feedback equalizer taps
EqDel           = ceil(EqNf/2);     % Equalizer delay
PAlloc          = 6.5;              % Allocated dispersion penalty (dBo)
Q0              = 7.03;             % BER = 10^(-12)
 
%% STEP 1 - Process waveform through simulated fiber channel %%%%%%%%%%%%%%%%%%%
%% Load input waveforms
XmitData    = load(TxDataFile);
yout0       = load(MeasuredWaveformFile);
PtrnLength  = length(XmitData);
TotLen      = PtrnLength*OverSampleRate;
Fgrid       = [-TotLen/2:TotLen/2-1].'/(PtrnLength*SymbolPeriod);

% *** New code calculating signal strength from waveform
if OMA_Method
    Estimate_OMA;
end

%% Process through fiber model. Fiber frequency response is normalized to 1 at DC.
for i=1:3
PCoefs      = FiberResp(i+1,:)';
ExpArg      = -j*2*pi*Fgrid;
Hsys        = exp(ExpArg * Delays') * PCoefs;
Hx          = fftshift(Hsys/sum(PCoefs));
yout        = real(ifft(fft(yout0).*Hx));
 
%% STEP 2 - Normalize OMA%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
yout        = (yout - SteadyZeroPower)/MeasuredOMA;
 
%% STEP 3 - Process signal through front-end antialiasing filter %%%%%%%%%%%%%%%%%%
%% Compute frequency response of front-end Butterworth filter
[b,a]       = butter(4, 2*pi*EFilterBW,'s');
H_r         = freqs(b,a,2*pi*Fgrid);
%% Process signal through front-end filter
yout        = real(ifft(fft(yout) .* fftshift(H_r)));
 
%% STEP 4 - Sample at rate 2/T %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
yout        = yout(1:OverSampleRate/2:end);
 
%% STEP 5 - Compute MMSE-DFE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The MMSE-DFE filter coefficients computed below minimize mean-squared error at the slicer input.
%% The derivation follows from the fact that the slicer input over the period of the data sequence can be
%% expressed as Z = (R+N)*W - X*[0 B]', where R and N are Toeplitz matrices constructed from the signal
%% and noise components, respectively, at the sampled output of the antialiasing filter, W is the feed-
%% forward filter, X is a Toeplitz matrix constructed from the input data sequence, and B is the feedback
%% filter. The computed W and B minimize the mean square error between the input to the slicer and the
%% transmitted sequence due to residual ISI and Gaussian noise.
%% Compute the noise autocorrelation sequence at the output of the front-end filter and rate-2/T sampler.
N0      = SymbolPeriod/(2 * Q0^2 * 10^(2*PAlloc/10));
Snn     = N0/2 * fftshift(abs(H_r).^2) * 1/SymbolPeriod * OverSampleRate;
Rnn     = real(ifft(Snn));
Corr    = Rnn(1:OverSampleRate/2:end);
%% Construct a Toeplitz autocorrelation matrix.
C       = toeplitz(Corr(1:EqNf));
%% Construct Toeplitz matrix from input data sequence
X       = toeplitz(XmitData, [XmitData(1); XmitData(end:-1:end-EqNb+1)]);
%% Construct Toeplitz matrix from signal at output of 2/T sampler.
%% This sequence gets wrapped by equalizer delay
R       = toeplitz(yout, [yout(1); yout(end:-1:end-EqNf+2)]);
R       = [R(EqDel+1:end,:); R(1:EqDel,:)];
R       = R(1:2:end, :);
%% Compute least-squares solution for filter coefficients
RINV    = inv(R'*R+PtrnLength*C);
P       = X'*(eye(PtrnLength) - R*RINV*R')*X;
P01     = P(1,2:EqNb+1);
P11     = P(2:EqNb+1,2:EqNb+1);
B       = -inv(P11)*P01'; % Feedback filter
W       = RINV*R'*X*[1;B]; % Feedforward filter
Z       = R*W - X*[0;B]; % Input to slicer
 
%% STEP 6 - Compute BER using semi-analytic method %%%%%%%%%%%%%%%%%%%%%%
MseGaussian = W'*C*W;
Ber         = sum(0.5*erfc((abs(Z-0.5)/sqrt(MseGaussian))/sqrt(2)))/length(Z);
 
%% STEP 7 - Compute equivalent SNR %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This function computes the inverse of the Gaussian error probability function. The built-in
%% function erfcinv() is not sensitive enough for the low probability of error case.
Q       = inf;
if Ber>10^(-12)      Q = sqrt(2)*erfinv(1-2*Ber);
elseif Ber>10^(-300) Q = 2.1143*(-1.0658-log10(Ber)).^0.5024;
end
 
%% STEP 8 - Compute penalty %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
RefSNR      = 10 * log10(Q0) + PAlloc;
TrialTWDP(i) = RefSNR-10*log10(Q);
end
 
%% Pick highest value due to the multiple responses from TrialTWDP.
TWDP    = max(TrialTWDP)
% End of program